@@ -3,37 +3,26 @@ from Crypto.Cipher import AES
 from Crypto.Util.Padding import pad, unpad
 from python_mpv_jsonipc import MPV
 from configparser import ConfigParser
-import socket
+
 import os
-import hashlib
 import sys
-import threading
-import errno
 import time
+import errno
+import socket
+import hashlib
 import datetime
 
 global connected
 global mpv
 
-
-
-
-
 def receive_message(client_socket):
 
     try:
-
-        # Receive our "header" containing message length, it's size is defined and constant
         message_header = client_socket.recv(HEADER_LENGTH)
         message_length = int(message_header)
 
     except:
-
-        # If we are here, client closed connection violently, for example by pressing ctrl+c on his script
-        # or just lost his connection
-        # socket.close() also invokes socket.shutdown(socket.SHUT_RDWR) what sends information about closing the socket (shutdown read/write)
-        # and that's also a cause when we receive an empty message
-        return False
+        return "Reading Error", "Server"
 
     msg = client_socket.recv(message_length)
 
@@ -45,21 +34,16 @@ def receive_message(client_socket):
         user = return_list[1]
 
         return msg, user
+
     else:
         return decrypted_msg, "Server"
 
-    return decrypted_msg
-
-    # If we received no data, client gracefully closed a connection, for example using socket.close() or socket.shutdown(socket.SHUT_RDWR)
-    #if not len(message_header):
-    #    return False
-
-    # Convert header to int value
-
 def pause_video(mpv):
     mpv.command("set_property","pause", True)
 
 def play_video(mpv):
+    global stop
+    stop = False
     mpv.command("set_property","pause", False)
 
 def toggle_play(mpv):
@@ -72,36 +56,21 @@ def toggle_play(mpv):
 def new_video(mpv, new_link):
     global t_playback
     if new_link is not None:
-        pause_video(mpv)
         mpv.play(new_link)
-        play_video(mpv)
     t_playback = 0
 
-def new_message():
-    thread = True
-    global connected
-    while thread == True:
-        msg = input(f'{my_username} > ')
-        if msg == "q":
-            thread = False
-            connected = False
-        if msg[:3] == "mpv":
-            print (msg)
-        else:
-            send(client_socket, msg)
-
 def send(clientsocket, msg):
     global KEY
     cipher = AES.new(KEY, AES.MODE_CBC)
+    if msg:
+        if type(msg) is not bytes:
+           msg = msg.encode("utf-8")
 
-    if type(msg) is not bytes:
-       msg = msg.encode("utf-8")
+        msg = encrypt_message(msg)
+        send_length = prepare_concatenation(len(msg))
 
-    msg = encrypt_message(msg)
-    send_length = prepare_concatenation(len(msg))
-
-    clientsocket.sendall(send_length)
-    clientsocket.sendall(msg)
+        clientsocket.send(send_length)
+        clientsocket.send(msg)
 
 def encrypt_message(msg):
     global KEY
@@ -131,13 +100,13 @@ def prepare_concatenation(msg):
     concat += b' ' * (HEADER_LENGTH - len(concat))
     return concat
 
-def ready_when_seeked(mpv):
+def ready_when_seeked(mpv, value):
     while True:
         seek_bool = mpv.command("get_property","seeking")
         if seek_bool == False:
             pause_video(mpv)
+            send(client_socket, f"ready {value}")
             break
-    send(client_socket, "ready")
 
 def exit_gracefully():
     send(client_socket, "!DISCONNECT")
@@ -147,54 +116,161 @@ def handle_server(server,addr):
     global t_playback
     global mpv
     global client_socket
+    global stop
+    global restart
 
     t_playback = 0
+    t_restart = 0
+    restart = 0
     connected = True
     mpv = MPV(start_mpv=True, quit_callback=exit_gracefully)
     mpv.command("set_property","osd-font-size","18")
 
+
+    @mpv.property_observer("paused-for-cache")
+    def observe_playback_time(name, value):
+        global restart
+        if value == True:
+            print("seek = true")
+            restart+=1
+            if restart == 1:
+                restart = 0
+                restart_url = mpv.command("get_property","path")
+                skip_time = mpv.command("get_property","playback-time")
+
+                send(client_socket, "toggle play")
+                mpv.play("https://www.youtube.com/watch?v=dQw4w9WgXcQ")
+               
+                mpv.play(restart_url)
+                #ready_when_seeked(mpv, value)
+                while True:
+                    seek_bool = mpv.command("get_property","seeking")
+                    if seek_bool == False:
+                        pause_video(mpv)
+                        break
+                
+                mpv.command("set_property","playback-time",skip_time)
+                #new_video(mpv, "https://www.youtube.com/watch?v=dQw4w9WgXcQ")
+                
+
+
+    # observe playback-time to synchronize when user skips on timeline 
+
+
+
     @mpv.property_observer("playback-time")
     def observe_playback_time(name, value):  
         global t_playback  
+        global stop
         if value is not None:
-            if value > t_playback+0.25 or value < t_playback:
-                t_playback = mpv.command("get_property", "playback-time")
-                pause_video(mpv)
-                print (t_playback)
-                send(client_socket, f"mpv skip {t_playback}")
+            if value > t_playback+0.25 or value < t_playback-0.1:
+
+                if f"mpv skip {value}" != msg:
+                    stop = True
+                    t_playback = mpv.command("get_property", "playback-time")
+                    print (t_playback)
+                    send(client_socket, f"mpv skip {t_playback}")
+                    ready_when_seeked(mpv, value)
+
             t_playback = value
     
+    # observe path to distribute new video url to other clients 
+
     @mpv.property_observer("path")
     def observe_path(name, value):
+        global stop
+        print(name, value)
         if value is not None:
-            print (f"New Path: {value}")            
-            send(client_socket, f"mpv new {value}")
-            new_video(mpv, value)
+            print (f"New Path: {value}")
+            print(msg)
+            if f"mpv new {value}" != msg:   
+                stop = True         
+                send(client_socket, f"mpv new {value}")
+                new_video(mpv, value)
+                print("READY - PATH OBSERVED")
+                ready_when_seeked(mpv, value)
+
+    # when space pressed inform other clients of play/pause
 
     @mpv.on_key_press("space")
     def toggle_playback():
-        toggle_play(mpv)
-        send(client_socket, "toggle play")
+        global stop
+        if stop == False:
+            toggle_play(mpv)
+            send(client_socket, "toggle play")
+
+    # when q is pressed exit gracefully
 
     @mpv.on_key_press("q")
     def terminate():
         global connected
         print("Q")
         send(client_socket, "!DISCONNECT")
-        #connected = False
+        connected = False
+
+    @mpv.on_key_press(",")
+    def frame_back_step():
+        send(client_socket, "frame-back-step")
+        pause_video(mpv)
+        mpv.command('frame-back-step')
+        print('nigga')
+
+    @mpv.on_key_press(".")
+    def frame_step():
+        send(client_socket, "frame-step")
+        #pause_video(mpv)
+        mpv.command('frame-step')
+        print('nigga')
+
+    @mpv.on_key_press("b")
+    def add_speed():
+        print("")
+        send(client_socket, "add_speed")
+        mpv.command('add','speed',0.1)
+        print('sped up')
+
+    @mpv.on_key_press("n")
+    def subtract_speed():
+        print("n")
+        send(client_socket, "subtract_speed")
+        mpv.command('add','speed',-0.1)
+        print('slowed down')
+
+    @mpv.on_key_press("r")
+    def resync():
+        time_pos = mpv.command("get_property","time-pos")
+        print(time_pos)
+
+        send(client_socket, f"mpv skip {time_pos}")
+        ready_when_seeked(mpv, time_pos)
+
+
 
     print(f"[CONNECTION ESTABLISHED] to {addr}")
     
     while connected:
+        global stop
+
         try:
             msg, user = receive_message(server)
 
-            if msg:
-
+            if msg != False:
                 if msg == "!DISCONNECT":
                     connected = False
                     break
 
+                if msg == "frame-step":
+                    mpv.command('frame-step')
+
+                if msg == 'add_speed':
+                    mpv.command('add','speed',0.1)          
+                    
+                if msg == 'subtract_speed':
+                    mpv.command('add','speed',-0.1)  
+
+                if msg == "frame-back-step":
+                    mpv.command('frame-back-step')
+
                 if msg == "mpv pause":
                     pause_video(mpv)
 
@@ -216,6 +292,7 @@ def handle_server(server,addr):
                     mpv.command("show-text",f"{user} toggles", "1500")
 
                 if "mpv skip" in msg:
+                    stop = True
                     t_playback = float(msg.split(" ")[2])
 
                     if t_playback < 3600:
@@ -225,35 +302,35 @@ def handle_server(server,addr):
 
                     mpv.command("set_property","playback-time",msg.split(" ")[2])
                     mpv.command("show-text",f"{user} skips to {converted_time}", "1500")
-
-                    pause_video(mpv)                 
-                    ready_when_seeked(mpv)
+               
+                    ready_when_seeked(mpv, t_playback)
 
                 if "mpv new" in msg:
+                    stop = True
                     videopath = msg[8:]
                     new_video(mpv, videopath)
                     mpv.command("show-text",f"{user}: {videopath}","1500")
-                    pause_video(mpv)
-                    ready_when_seeked(mpv)
+                    ready_when_seeked(mpv, videopath)
+
+                if "userconnected" in msg:
+                    mpv.command("show-text",f"{msg.split(' ')[1]} connected.","1500")
+
 
         except IOError as e:
-        # This is normal on non blocking connections - when there are no incoming data error is going to be raised
-        # Some operating systems will indicate that using AGAIN, and some using WOULDBLOCK error code
-        # We are going to check for both - if one of them - that's expected, means no incoming data, continue as normal
-        # If we got different error code - something happened
             if e.errno != errno.EAGAIN and e.errno != errno.EWOULDBLOCK:
                 print(e)
                 print('Reading error: {}'.format(str(e)))
-                sys.exit()
+                continue
+                #sys.exit()
 
             # We just did not receive anything
-                continue
+                #break
 
         except Exception as e:
-            # Any other exception - something happened, exit
+            # Any other exception - something happened
             print(e)
             print('Reading error: '.format(str(e)))
-            sys.exit()
+            #sys.exit()
 
     mpv.terminate()
     client_socket.shutdown(socket.SHUT_RDWR)
@@ -328,22 +405,16 @@ def main():
 
     config = ConfigParser()
     config.read(configfile)
-    config.set('connection','ip', '%s'%IP)
+    config.set('connection','ip', '%s' % IP)
 
-    with open(configfile,"w") as f:
+    with open(configfile, "w") as f:
         config.write(f)
 
-    # Set connection to non-blocking state, so .recv() call won;t block, just return some exception we'll handle
     client_socket.setblocking(True)
 
-    # Prepare username and header and send them
-    # We need to encode username to bytes, then count number of bytes and prepare header of fixed size, that we encode to bytes as well
-
     send(client_socket, USERNAME)
 
-    handle_server(client_socket, (IP,PORT))
-# Create a socket
-# socket.AF_INET - address family, IPv4, some otehr possible are AF_INET6, AF_BLUETOOTH, AF_UNIX
-# socket.SOCK_STREAM - TCP, conection-based, socket.SOCK_DGRAM - UDP, connectionless, datagrams, socket.SOCK_RAW - raw IP packets
+    handle_server(client_socket, (IP, PORT))
+
 if __name__ == "__main__":
     main()
